"""\
Tools for storing form data for multi-page forms known as Wizards.

"""

import datetime
import logging
import md5   
import random

try:
    import formencode
except ImportError:
    formencode_present = False
else:
    formencode_present = True

from bn import AttributeDict
from conversionkit import Conversion, chainConverters
from nestedrecord import decodeNestedRecord, encode_error


log = logging.getLogger(__name__)
ERROR_STATUS = 400

# It is important that each validator has filter_extra_fields=True
# so that if we use the merge, a malicious user can't get variables
# into a future step.

def model_wizard_update(flow, wizard_id, values):
    raise Exception('Not tested yet!')
    flow.database.update_record(
        'wizard',
        'wizard_id',
        wizard_id,
        values
    )
    
def model_wizard_migrate(flow, from_wizard_id, to_wizard_id):
    # Get the old wizard's contents
    from_fields = model_wizard_dump_fields(flow, from_wizard_id)
    # Overwrite any data in the new wizard with the old data
    model_wizard_insert(flow, to_wizard_id, from_fields)
    # Now expire the old wizard
    flow.database.update_record(
        'wizard',
        'wizard_id',
        from_wizard_id,
        dict(
            status=wizard_status.MIGRATED,
        ),
    )
    # Now delete all the old keys
    model_wizard_delete(flow, from_wizard_id)
    
def model_wizard_dump_fields(flow, wizard_id, types=None):
    sql = 'SELECT type, key, value FROM wizard_field WHERE wizard_id=%s'
    values = [wizard_id]
    if types is not None:
        sql += 'AND type IN ('+(('%s, '*len(types))[:-2])+')'
        for value in types:
            values.append(value)
    return flow.database.query(sql, tuple(values), format='list')

def model_wizard_get_fields(flow, wizard_id, types=None):
    sql = 'SELECT key, value FROM wizard_field WHERE wizard_id=%s'
    values = [wizard_id]
    if types is not None:
        sql += 'AND type IN ('+(('%s, '*len(types))[:-2])+')'
        for value in types:
            values.append(value)
    result = {}
    rows = flow.database.query(sql, tuple(values), format='list')
    for row in rows:
        if result.has_key(row[0]):
            raise Exception(
                'Duplicate key %s found in wizard %s'%(row[0], wizard_id)
            )
        result[row[0]] = row[1]
    return result
    
def model_wizard_complete(flow, wizard_id):
    """\
    Complete the wizard, removing all the old keys but leaving the accessed
    time untocuhed.
    """
    flow.database.update_record(
        'wizard',
        'wizard_id',
        wizard_id,
        dict(
            status=wizard_status.COMPLETED,
        ),
    )
    # Now delete all the old keys
    model_wizard_delete(flow, wizard_id)

def model_wizard_expire(flow, wizard_id):
    """\
    Expire the wizard, removing all the old keys but leaving the accessed 
    time untocuhed.
    """
    flow.database.update_record(
        'wizard',
        'wizard_id',
        wizard_id,
        dict(
            status=wizard_status.EXPIRED,
        ),
    )
    # Now delete all the old keys
    model_wizard_delete(flow, wizard_id)

def generate_hash(flow, id, secret):
    hash = md5.md5()
    hash.update(str(id))
    hash.update(secret)
    hash = hash.hexdigest()[3:9]
    return hash

wizard_status = AttributeDict(
    ACTIVE=1,
    CAPTCHA_NEEDED=2,
    EXPIRED=3,
    COMPLETED=4,
    HACKED=5,
    MIGRATED=6,
)

wizard_field_type = AttributeDict(
    PUBLIC=0,
    PRIVATE=1,
)

def model_wizard_new(
    flow,
    name,
    expires=432000, # 5 Days
    status=wizard_status['ACTIVE'],
    accessed=None,
    secret=None,
    add_to_flow=True,
):
    if accessed is None:
        accessed = datetime.datetime.now()
    wizard_id = flow.database.insert_record(
        'wizard',
        dict(
            name=name,
            expires=expires,
            status=status,
            # accessed could be automatically set by the database as a default
            # but then we couldn't return it below
            accessed=accessed,
            failed_attempts=0,
        ),
        'wizard_id',
    )
    # Now update it with the hash used so that even if the secret changes
    # existing wizards will continue to function
    if secret is None:
        secret = flow.config.wizard.secret
    hash = generate_hash(flow, wizard_id, secret)
    flow.database.update_record(
        'wizard',
        'wizard_id',
        wizard_id,
        dict(
            hash=hash
        ),
    )
    data = AttributeDict(
        wizard_id=wizard_id,
        hash=hash,
        name=name,
        expires=expires,
        status=status,
        accessed=accessed,
        failed_attempts=0,
    )
    if add_to_flow:
        flow['wizard'] = data
    return data

def model_wizard_delete(flow, wizard_id, later_steps=None, types=None):
    sql = "DELETE FROM wizard_field WHERE wizard_id=%s"
    tup = []
    if later_steps is not None:
        sql += " AND ("+(' key LIKE(%s) OR '*len(later_steps))[:-3]+')'
        tup = [step+'.%' for step in later_steps]
    tup.insert(0, wizard_id)
    if types is not None:
        sql += ' AND type IN ('+(('%s, '*types)[:-2])+')'
        for value in types:
            tup.append(value)
    flow.database.query(
       sql,
       tuple(tup),
       fetch=False
    )

def model_wizard_insert(flow, wizard_id, values):
    connection = flow.database.connect()
    cursor = connection.cursor()
    vs = []
    for value in values:
        vs.append((wizard_id, value[0], value[1], value[2]))
    cursor.executemany(
        "INSERT INTO wizard_field (wizard_id, type, key, value) VALUES (%s, %s, %s, %s)",
        tuple(vs)
    )

def model_wizard_insert_private(flow, wizard_id, values):
    vs = []
    for k, v in values.items():
        vs.append((wizard_field_type['PRIVATE'], k, v))
    return model_wizard_insert(flow, wizard_id, vs)

def model_wizard_insert_public(flow, wizard_id, values):
    vs = []
    for k, v in values.items():
        vs.append((wizard_field_type['PUBLIC'], k, v))
    return model_wizard_insert(flow, wizard_id, vs)

def model_wizard_purge_expired(flow):
    connection = flow.database.connect()
    rows = flow.database.query(
        """
        SELECT wizard_id FROM wizard 
        WHERE accessed+(expires*'1 second'::interval) < NOW()
        AND status=%s
        """,
        (wizard_status['ACTIVE'],),
        format='list',
    )
    delete = ''
    values = []
    if rows:
        for row in rows:
            delete+='%s, '
            values.append(row[0])
        cursor = connection.cursor()
        cursor.execute(
           """
           DELETE FROM wizard_field
           WHERE wizard_id in (%s)
           """%delete[:-2],
           tuple(values)
        )
        terms = values.copy()
        terms.insert(0, wizard_status['EXPIRED'])
        cursor.execute(
           """
           UPDATE wizard
           SET status=%%s
           WHERE wizard_id in (%s)
           """%delete[:-2],
           tuple(terms)
        )
        return len(values)

class WizardCaptchaNeeded(Exception):
    def __init__(self, wizard):
        self.wizard = wizard

class InvalidWizard(Exception):
    def __init__(self, wizard):
        self.wizard = wizard

class NoWizardInformation(Exception):
    pass

class InvalidHash(InvalidWizard):
    pass

class Expired(InvalidWizard):
    pass

class Completed(InvalidWizard):
    pass

class NoSuchWizard(Exception):
    pass

def model_wizard_get_and_check(flow, wizard_id, hash, allowed_failures):
    rows = flow.database.query(
        "SELECT * FROM wizard WHERE wizard_id=%s",
        (int(wizard_id),)
    )
    if len(rows) > 1:
        raise Exception('Too many rows returned from the wizard table')
    elif not rows:
        raise NoSuchWizard()
    else:
        wizard = AttributeDict(rows[0])
        if wizard['accessed'] + datetime.timedelta(
           seconds=wizard['expires']) < datetime.datetime.now():
            # The wizard has expired
            # Don't update the accessed_counter
            flow.database.update_record(
                'wizard',
                'wizard_id',
                wizard_id,
                dict(
                    failed_attempts=wizard['failed_attempts']+1,
                    status=wizard_status['EXPIRED']
                ),
            )
            raise Expired(wizard)
        elif int(wizard['failed_attempts']) >= int(allowed_failures):
            flow.database.update_record(
                'wizard',
                'wizard_id',
                wizard_id,
                dict(
                    failed_attempts=wizard['failed_attempts']+1,
                    status=wizard_status['CAPTCHA_NEEDED']
                ),
            )
            raise WizardCaptchaNeeded(wizard)

        if wizard['status'] != wizard_status['ACTIVE'] or \
           wizard['hash'].strip() != hash.strip():
            # Just increment failed attempts
            flow.database.update_record(
                'wizard',
                'wizard_id',
                wizard_id,
                dict(
                    failed_attempts=wizard['failed_attempts']+1,
                ),
            )
        else:
            # Update the accessed_counter
            accessed = datetime.datetime.now()
            flow.database.update_record(
                'wizard',
                'wizard_id',
                wizard_id,
                dict(
                    accessed=accessed
                ),
            )
            wizard[accessed] = accessed
        # Now worry about raising the correct exception or returning the result
        if wizard['status'] == wizard_status['CAPTCHA_NEEDED']:
            raise WizardCaptchaNeeded(wizard)
        elif wizard['hash'].strip() != hash.strip():
            raise InvalidHash(wizard)
        elif wizard['status'] == wizard_status['COMPLETED']:
            raise Completed(wizard)
        return wizard

#
# ConversionKit implementation
#

def CKRemoveLaterSteps(steps):
    """\
This pre_validator needs to:
    * remove any fields associated with steps later than the current one and purge them from the wizard store
    * ensure that values for each of the steps prior to the current one are validated
    * The whole wizard is treated as a single namespace so there is no separation into steps. Which field is treated as being associated with which step is entirely up to the validators. As long as they validate it doesn't matter what extra fields are passed to them or from which step. This means when the wizard is being refactored, data from subsequent steps can still populate fields from earlier steps.
    """
    def remove_empty_steps_converter(conversion, state):
        final = {}
        need_purge = False
        for k, v in conversion.value.items():
            for step in steps:
                if k.startswith(step+'.') or k == step:
                    final[k] = v
        conversion.result = final
    return remove_empty_steps_converter

from httpkit.helper.cookie import get_cookie

def wizard_from_cookie(cookie_name):
    """\
    A decorator which takes a cookie name and extracts the wizard_id and hash
    from the cookie value and ensures they are both valid.

    The cookie value is in the form wizard_id+wizard_hash
    """
    def get_keys(flow):
        cookie = get_cookie(flow.wsgi.environ.get('HTTP_COOKIE'), cookie_name)
        if cookie is None:
            raise NoWizardInformation(None)
        return cookie.split('+')
    return wizard_decorator(get_keys)

def wizard_from_session(wizard_key, hash_key):
    """\
    A decorator which takes a wizard ID and a hash, extracts them from the
    session and ensures they are both valid.
    """
    def get_keys(flow):
        if not (flow.session.has_key(wizard_key) and \
           flow.session.has_key(hash_key)):
            raise NoWizardInformation(None)
        return (flow.session[wizard_key], flow.session[hash_key])
    return wizard_decorator(get_keys)

def wizard_from_input(wizard_key, hash_key):
    """\
    A decorator which takes a wizard ID and a hash, extracts them from query
    parameters and ensures they are both valid.
    """
    def get_keys(flow):
        if not flow.http.input or not flow.http.input.has_key(wizard_key) or not flow.http.input.has_key(hash_key):
            raise NoWizardInformation(None)
        return (flow.http.input.getfirst(wizard_key), flow.http.input.getfirst(hash_key))
    return wizard_decorator(get_keys)

def wizard_from_query(wizard_key, hash_key):
    """\
    A decorator which takes a wizard ID and a hash, extracts them from query
    parameters and ensures they are both valid.
    """
    def get_keys(flow):
        if not flow.wsgi.environ.get('QUERY_STRING') or not (flow.query.has_key(wizard_key) and \
           flow.query.has_key(hash_key)):
            raise NoWizardInformation(None)
        return (flow.query.getfirst(wizard_key), flow.query.getfirst(hash_key))
    return wizard_decorator(get_keys)

def wizard_from_urlvars(wizard_key, hash_key):
    """\
    A decorator which takes a wizard ID and a hash, extracts them from the URL
    and ensures they are both valid
    """
    def get_keys(flow):
        if not (flow.url.vars.has_key(wizard_key) and \
           flow.url.vars.has_key(hash_key)):
            raise NoWizardInformation(None)
        return (
            flow.url.vars[wizard_key], 
            flow.url.vars[hash_key],
        )
    return wizard_decorator(get_keys)

def wizard_decorator(get_keys):
    def wizz(func):
        def handle(flow):
            if str(random.random()) < \
               flow.config.wizard.expire_probability:
                model_wizard_purge_expired(flow)
            try:
                wizard_id, hash = get_keys(flow)
            except NoWizardInformation:
                # Return the error page.
                log.error(
                    'No wizard information could be found, requested by %s',
                    flow.wsgi.environ.get(
                        'REMOTE_ADDR', 
                        flow.wsgi.environ.get('HTTP_X_FORWARDED_FOR')
                    ),
                )
                if flow.config.app.debug:
                    raise
                else:
                    return flow.url.abort(ERROR_STATUS)
            else:
                try:
                    wizard = model_wizard_get_and_check(
                        flow,
                        wizard_id=wizard_id,
                        hash=hash,
                        allowed_failures=flow.config.wizard.allowed_failures,
                    )
                except (NoSuchWizard), e:
                    # Return the error page.
                    log.error(
                        'The wizard %s %s requested by %s could not be found',
                        wizard_id, 
                        hash,
                        flow.wsgi.environ.get(
                            'REMOTE_ADDR', 
                            flow.wsgi.environ.get('HTTP_X_FORWARDED_FOR')
                        )
                    )
                    return flow.url.abort(ERROR_STATUS)
                except (InvalidWizard), e:
                    # Return the error page.
                    log.error(
                        'The wizard %s %s requested by %s was invalid',
                        wizard_id, 
                        hash,
                        flow.wsgi.environ.get(
                            'REMOTE_ADDR', 
                            flow.wsgi.environ.get('HTTP_X_FORWARDED_FOR')
                        )
                    )
                    if flow.wsgi.config.app.debug:
                        raise
                    else:
                        return flow.url.abort(ERROR_STATUS)
                except WizardCaptchaNeeded, e:
                     wizard = e.wizard
                     # Now redirect to the recaptcha screen.
                     raise Exception('Not implemented yet')
                     flow.url.redirect_to(
                         controller='wizard', 
                         action='index', 
                         wizard_id=wizard['wizard_id'],
                     )
                else:
                    flow['wizard'] = AttributeDict(wizard)
                    return func(flow)
        return handle
    return wizz

def wizard_validate_steps(steps, validator, start_dict):
    def wizz(func):
        def handler(flow):
            # Get data for all steps
            wizard_cache = model_wizard_get_fields(
                flow,
                flow.wizard.wizard_id,
                types=[wizard_field_type['PUBLIC'], wizard_field_type['PRIVATE']]
            )
            flow.wizard['field'] = flow.wizard['value'] = wizard_cache
            if formencode_present and isinstance(validator, (formencode.validators.FancyValidator, formencode.Schema)):
                # Verify data from core_info 
                validation_state=create_validation_state(
                    flow,
                    steps
                )
                try:
                    result = validator.to_python(wizard_cache, validation_state)
                except formencode.Invalid, e:
                    errors = formbuild.errors_to_dict(e)
                    log.error('Data from the previous steps is invalid: %r', errors)
                    return 'Invalid data, please go back or <a href="'+\
                       flow.url.url_for(**start_dict)+'">start again</a>.'
                else:
                    flow.wizard['result'] = result
                    return func(flow)
            else:
                converter = chainConverters(CKRemoveLaterSteps(steps), decodeNestedRecord(depth=1), validator)
                conversion = Conversion(wizard_cache).perform(converter, flow)
                if conversion.successful:
                    flow.wizard['result'] = conversion.result
                    return func(flow)
                else:
                    raise Exception('Invalid wizard: %r', encode_error(conversion))
        return handler
    return wizz

